#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "global.h"
#include "qmail.h"
#include "auto_qmail.h"
#include "stralloc.h"
#include "case.h"
#include "control.h"
#include "constmap.h"
#include "direntry.h"
#include "error.h"
#include "getoptb.h"
#include "str.h"
#include "fmt.h"
#include "open.h"
#include "byte.h"
#include "scan.h"
#include "str.h"
#include "error.h"

#define FDAUTH 3

/** @file qmail-vmailuser.c
    @return 0: virtual user exists 
            1: virtual user dir not accessible
            2: qmail-vmailuser is misused
            110: can't read controls
            111: temporary problem
*/

char inputbuf[BUFSIZE_AUTH];
struct constmap mapvdoms;
stralloc vdoms = {0};
stralloc vdomdir = {0};
stralloc vuser = {0};
stralloc vuserdir = {0};

int pam_exit(int fail)
{
  int i;

  close(FDAUTH);
  for (i = 0; i < sizeof(inputbuf); ++i) inputbuf[i] = 0;
  _exit(fail);
}

int main(int argc,char * const *argv)
{
  DIR *dir;
  char *vdomuser; 
  char *domain = 0;
  int buflen = 0;
  int domlen = 0;
  int flagrespect = 0;
  int flagvpopmail = 1;
  int flagvmailmgr = 1;
  int flagunixuser = 1;
  int opt;
  int r;
  char ch;
  char *homedir = "/home";

  while ((opt = getoptb(argc,(char **)argv,"cCMPU")) != opteof)
    switch (opt) {
      case 'c': flagrespect = 0; break;
      case 'C': flagrespect = 1; break;
      case 'M': flagvpopmail = 0; flagunixuser = 0; break; // only vmailmgr
      case 'P': flagunixuser = 0; flagvmailmgr = 0; break; // only vpopmail
      case 'U': flagvpopmail = 0; flagvmailmgr = 0; break; // only unix virtual user
    }
  argv += optind;
  argc -= optind;

  if (argv && *argv) {
    homedir = *argv;
    dir = opendir(homedir);
    if (!dir) pam_exit(2);
  }

  if (chdir(auto_qmail) == -1) pam_exit(110);
 
  switch (control_readfile(&vdoms,"control/virtualdomains",0)) {
    case -1: pam_exit(110);
    case  0: if (!constmap_init(&mapvdoms,"",0,1)) pam_exit(111);
    case  1: if (!constmap_init(&mapvdoms,vdoms.s,vdoms.len,1)) pam_exit(111);
  }

  for (;;) {						/* read input */
    do
      r = read(FDAUTH,inputbuf + buflen,sizeof(inputbuf) - buflen);
    while ((r == -1) && (errno == EINTR));
    if (r == -1) pam_exit(111);
    if (r == 0) break;
    buflen += r;
    if (buflen >= sizeof(inputbuf)) pam_exit(2);
  }
  close(FDAUTH);

  if ((r = byte_rchr(inputbuf,buflen,'@')))			 /* @domain */
    if (r < buflen && inputbuf[r] == '@') {
      domain = inputbuf + r + 1;
      domlen = str_len(domain);
      if (!flagrespect)
        case_lowerb(inputbuf,buflen);
      else
        case_lowerb(domain,domlen);
    }
  vdomuser = constmap(&mapvdoms,domain,domlen);
  if (!vdomuser) pam_exit(1);

  if (!stralloc_copys(&vuser,"")) pam_exit(111); /* user */
  for (int i = 0; i < r; ++i) {
    ch = inputbuf[i];
    if (ch == '.') ch = ':';
    if (!stralloc_append(&vuser,&ch)) pam_exit(111);
  }
  if (!stralloc_0(&vuser)) pam_exit(111); 

  if (!case_diffs(vdomuser,vuser.s)) pam_exit(1);  /* trivial case */

  if (flagvpopmail) {                            /* vpopmail */
    if (!stralloc_copys(&vdomdir,homedir)) pam_exit(111);  
    if (!stralloc_cats(&vdomdir,"/")) pam_exit(111);  
    if (!stralloc_cats(&vdomdir,"vpopmail")) pam_exit(111);
    if (!stralloc_copy(&vuserdir,&vdomdir)) pam_exit(111);  
    if (!stralloc_cats(&vuserdir,"/domains/")) pam_exit(111);
    if (!stralloc_cats(&vuserdir,vdomuser)) pam_exit(111);  
    if (!stralloc_cats(&vuserdir,"/")) pam_exit(111);  
    if (!stralloc_copy(&vdomdir,&vuser)) pam_exit(111);  
    if (!stralloc_0(&vdomdir)) pam_exit(111); 

    dir = opendir(vdomdir.s);
    if (dir) {
      if (!stralloc_cats(&vuserdir,"/")) pam_exit(111);  
      if (!stralloc_cats(&vuserdir,vuser.s)) pam_exit(111);  
      if (!stralloc_0(&vuserdir)) pam_exit(111); 

      dir = opendir(vuserdir.s);
      if (dir && !errno) pam_exit(0);
    }
    pam_exit(1);
  }

  if (flagvmailmgr) {                            /* vmailmgr */
    if (!stralloc_copys(&vdomdir,homedir)) pam_exit(111);  
    if (!stralloc_cats(&vdomdir,"/")) pam_exit(111);  
    if (!stralloc_copy(&vuserdir,&vdomdir)) pam_exit(111);  
    if (!stralloc_cats(&vuserdir,vdomuser)) pam_exit(111);  
    if (!stralloc_cats(&vuserdir,"/users/")) pam_exit(111);  
    if (!stralloc_copy(&vdomdir,&vuserdir)) pam_exit(111);  
    if (!stralloc_0(&vdomdir)) pam_exit(111);  // /home/vdomuser/users/ 

    dir = opendir(vdomdir.s);
    if (dir) {
      if (!stralloc_cats(&vuserdir,vuser.s)) pam_exit(111);  
      if (!stralloc_0(&vuserdir)) pam_exit(111); 

      dir = opendir(vuserdir.s);
      if (dir) pam_exit(0);
    } 
    pam_exit(1);
  }

  if (flagunixuser) {                            /* standard virtual unix user */
    if (!stralloc_copys(&vdomdir,homedir)) pam_exit(111);  
    if (!stralloc_cats(&vdomdir,"/")) pam_exit(111);  
    if (!stralloc_cats(&vdomdir,vdomuser)) pam_exit(111);  
    if (!stralloc_0(&vdomdir)) pam_exit(111); 

    dir = opendir(vdomdir.s);
    if (dir) pam_exit(0);
  }
   
  return pam_exit(1);
}
